import logging
from typing import List, Optional, Dict, Any
from datetime import datetime, timedelta

default_logger = logging.getLogger(__name__)


class QueryConfig:
    def __init__(
        self, name: str, query: str, value_columns: List[str], dimension_columns: List[str], will_run_now: bool = False, extra_dimensions: Optional[str] = None
    ):
        super().__init__()
        self.name: str = name
        self.query: str = query
        self.value_columns: List[str] = value_columns
        self.dimension_columns: List[str] = dimension_columns
        self.will_run_now = will_run_now
        self.extra_dimensions = extra_dimensions

    def __repr__(self):
        return f"Query(name: {self.name}, value_columns:{self.value_columns}, dimension_columns:{self.dimension_columns}, extra_dimensions={self.extra_dimensions})"


class Column(object):
    def __init__(self, col_json: dict):
        self.name = col_json.get("name")
        self.value = col_json.get("value")
        self.index = col_json.get("index")

    def __str__(self):
        return f"{self.name}: {self.value}"

    def __repr__(self):
        return f"Column(Index:{self.index},: Name={self.name}, Value={self.value})"


class Row(object):
    def __init__(self, columns: List[Column]):
        self.columns = sorted(columns, key=lambda col: col.index)

    def __repr__(self):
        return f"Row(columns={self.columns})"

    def get_column_by_index(self, index) -> Column:
        for c in self.columns:
            if c.index == index:
                return c
        raise Exception(f"Index {index} is out of range ({len(self.columns)} columns present)")

    def get_column_by_name(self, name: str) -> Column:
        for c in self.columns:
            if str(c.name).upper() == name.upper():
                return c
        raise Exception(f"Could not find column with name: '{name}'. Only had: {[c.name for c in self.columns]}")


class Result:
    def __init__(self):
        self.error: bool = False
        self.error_message = ""
        self.query_results: List[QueryResult] = []
        self.duration: int = 0
        self.timestamp: int = 0

    def human_timestamp(self) -> str:
        timestamp = datetime.utcfromtimestamp(self.timestamp)
        time_ago = datetime.utcnow() - timestamp
        return f"{timestamp} ({time_ago.seconds / 60:.1f} minutes ago)"

    @staticmethod
    def from_json(raw_data: Dict[str, Any]) -> "Result":
        r = Result()
        r.error = raw_data.get("error", False)
        r.duration = raw_data.get("duration", 0)
        r.timestamp = raw_data.get("timestamp", 0)
        r.error_message = raw_data.get("errorMessage", "")

        for query_result in raw_data.get("queries", []):
            r.query_results.append(QueryResult.from_json(query_result))

        return r

    def __repr__(self):
        return f"Result(error={self.error}, timestamp={self.human_timestamp()}, duration={self.duration}, error_message={self.error_message}, queries={len(self.query_results)})"


class QueryResult(object):
    def __init__(self, logger=default_logger):
        self.error: bool = False
        self.error_message: str = ""
        self.rows: List[Row] = []
        self.logger = logger
        self.name: str = ""
        self.duration: int = 0
        self.timestamp: int = 0

    @staticmethod
    def from_json(json_dict: dict) -> "QueryResult":
        r = QueryResult()
        if json_dict is not None:
            try:
                r.error = json_dict["error"]
                r.error_message = json_dict.get("errorMessage", "")
                r.duration = json_dict.get("duration", 0)
                r.timestamp = json_dict.get("timestamp", 0)
                r.name = json_dict.get("name", 0)

                if not r.error:
                    for row in json_dict.get("rows", []):
                        r.rows.append(Row([Column(col) for col in row["columns"]]))
            except Exception as e:
                error = f"Could not parse result: {e}. The result was: '{json_dict}'"
                r.error = True
                r.error_message = error
                r.logger.exception(error)

        return r

    def __repr__(self):
        return f"Result(name='{self.name}', error={self.error}, error_message='{self.error_message}', rows={len(self.rows)})"

    def __str__(self):
        return self.__repr__()
