import json
import os
import tempfile
import socket
import string
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Any, Optional, Callable
from datetime import datetime

from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.exceptions import NothingToReportException, ConfigException
from ruxit.api.topology_builder import Device

from mssql_imports import connector
from mssql_imports.log_proxy import LogProxy
from mssql_imports.models import Result, QueryResult
from mssql_imports.queries import queries

from mssql_imports.models import QueryConfig


class RemoteMsSqlPlugin(RemoteBasePlugin):

    def initialize(self, **kwargs):
        # Validate some user parameters
        self.connection_string = self.build_connection_string()
        if not self.connection_string.startswith("jdbc:sqlserver://"):
            raise ConfigException("Invalid connection string, it must start with jdbc:sqlserver://")

        self.temp_folder = create_or_get_temp_folder()
        self.metrics_by_query_name: Dict[str, List[dict]] = self.metrics_by_query_name()

        # Write all queries to the config file
        instance_filter = "instance_name != '_Total'"
        if self.config.get("single_database"):
            instance_filter = f"instance_name = '{self.get_database_name()}'"
        for query_config in queries:
            query_config.query = query_config.query.format(
                instance_filter=instance_filter,
            )
        self.write_config(queries)

    def query(self, **kwargs):
        start_time = datetime.now()
        self.logger.setLevel(self.config.get("log_level", "INFO"))
        self.custom_device = self.create_custom_device()
        self.report_endpoint()

        # Spawn the java process that will run the queries
        # The password is passed as an Environment Variable for security
        env = os.environ.copy()
        env["DATABASE_PASSWORD"] = self.config["database_password"]
        connector.query(self.activation.entity_id, self.temp_folder, env=env, java_path=self.config.get("jvm_path"))

        # Parse the results of the queries
        result = self.read_results_from_file()
        self.logger.info(f"Processing result {result}")
        total_reported_metrics = 1

        if result.error:
            # This means there was an error connecting to the Database
            # Report the error and the availability
            error_message = f"Error running queries: {result.error_message}"
            self.logger.error(error_message)
            self.custom_device.report_error_event(error_message, "MSSQL Extension error")
            self.custom_device.absolute("AvailabilityPercent", 0)

        else:
            # If we get here we have successfully queried the database, parse the results
            self.custom_device.absolute("AvailabilityPercent", 100)
            for query_result in result.query_results:
                self.logger.info(f"Processing query result {query_result}")

                # Grab all the metrics that depend on this query from plugin.json
                metrics_configs = self.metrics_by_query_name.get(query_result.name, [])
                total_reported_metrics += self.report_metrics(metrics_configs, query_result)

        self.parse_logs()
        time_taken = (datetime.now() - start_time).total_seconds()
        self.logger.info(
            f"Reported {total_reported_metrics} metrics to Dynatrace, query method time: {time_taken:.2f}s")

    def report_metrics(self, metric_configs: List[dict], query_result: QueryResult) -> int:
        total_reported_metrics = 0
        query_config = next((q for q in queries if q.name == query_result.name))
        for metric_config in metric_configs:
            try:
                total_reported_metrics += self.process_metric(metric_config, query_result, query_config)
            except Exception as e:
                error_message = f"Could not process metric {metric_config}: {e}"
                self.logger.exception(error_message)
                self.custom_device.report_custom_info_event(error_message, "MSSQL Extension - Error processing metric")

        return total_reported_metrics

    def process_metric(self, metric_config: dict, query_result: QueryResult, query_config: QueryConfig) -> int:
        total_reported_metrics = 0

        # First lets figure out what is the metric key
        # If we have dimensions, a formula and a metric_type
        # This list has already been filtered for only metrics that match the query name from query_result
        # All of this comes from plugin.json

        # Use the correct key depending on it being a normal or a state metric
        timeseries_key = "timeseries" if "timeseries" in metric_config else "statetimeseries"
        key = metric_config[timeseries_key]["key"]

        # Check if the user chose to monitor this metric or not (checkbox in the UI)
        should_monitor = self.config.get(key, True)
        if not should_monitor:
            self.logger.info(f"Skipping metric {key} because the user chose not to monitor it")
            return 0

        expected_dimensions = metric_config[timeseries_key].get("dimensions", [])
        formula = metric_config.get("source", {}).get("formula", "")  # This exists for every metric, checks just to be safe
        metric_type = metric_config.get("source", {}).get("type", "absolute")
        self.logger.debug(f"Processing metric {key} with formula: '{formula}' and type '{metric_type}'")

        # Trick to extract fields from a format string
        # From a string like '{rows_read} / ({rows_returned} + 1)'
        # This will generate ['rows_read', 'rows_returned']
        formula_names = [v[1] for v in string.Formatter().parse(formula) if v[1] is not None]

        # This is ugly, if we were using mint we could simply report the dimensions and values
        # But in this case we have to extract the dimensions we need for this metric that was
        # defined in plugin.json
        for row in query_result.rows:
            dimensions = {}
            for dimension_column in query_config.dimension_columns:
                column_value = row.get_column_by_name(dimension_column).value
                if column_value in formula_names:
                    # This is the row that contains our metric, extract values and dimensions

                    # Extracting dimensions
                    for expected_dimension in expected_dimensions:
                        dimensions[expected_dimension] = row.get_column_by_name(expected_dimension).value

                    # Extracting metrics
                    for value_column in query_config.value_columns:
                        # This loop only happens once for now as we have a single value per metric
                        value = row.get_column_by_name(value_column).value
                        self.report_metric(key, value, dimensions, metric_type)
                        total_reported_metrics += 1

        return total_reported_metrics

    def report_endpoint(self):
        try:
            ip = socket.gethostbyname(self.get_hostname())
            self.custom_device.add_endpoint(ip, int(self.get_port()), dnsNames=[self.get_hostname()])
        except Exception as e:
            self.logger.warning(f"Could not determine the host IP address: {e} to report an endpoint")

    def get_database_name(self) -> str:
        database_name = self.config.get("database_name")
        if self.config.get("connection_string"):
            database_name = self.config["connection_string"].split(";databaseName=")[1].split(";")[0]
        return database_name

    def get_hostname(self) -> str:
        hostname = self.config.get("database_host")
        if self.config.get("connection_string"):
            hostname = self.config["connection_string"].split("://")[1].split(":")[0]
        return hostname

    def get_port(self) -> str:
        port = self.config.get("database_port")
        if self.config.get("connection_string"):
            port = self.config["connection_string"].split("://")[1].split(":")[1].split(";")[0]
        return port

    def create_custom_device(self) -> Device:
        device_names = {
            "Database Name": self.get_database_name(),
            "Host": self.get_hostname(),
            "Host:Port": f"{self.get_hostname()}:{self.get_port()}",
            "Custom": self.config["custom_device_identifier"]
        }
        group_names = {
            "Database Name": self.get_database_name(),
            "Host": self.get_hostname(),
            "Host:Port": f"{self.get_hostname()}:{self.get_port()}",
            "Custom": self.config["custom_device_group_identifier"]
        }
        device_name = device_names[self.config["custom_device_identifier_source"]]
        group_name = group_names[self.config["custom_device_group_identifier_source"]]

        self.logger.info(f"Group name: '{group_name}', device name: '{device_name}'")

        group = self.topology_builder.create_group(group_name, group_name)
        return group.create_device(device_name)

    def metrics_by_query_name(self) -> Dict[str, List[dict]]:
        # Here we group the metric configuration by query name
        # We need this to later resolve formulas and report type (absolute, relative, per_second)
        results = defaultdict(list)
        for metric in self.plugin_info.json_data.get("metrics", []):
            query_name = metric.get("source", {}).get("queryName", "")
            results[query_name].append(metric)
        return results

    def build_connection_string(self) -> str:

        # Check if the user supplied a connection string
        connection_string = self.config.get("connection_string")
        if not connection_string:
            # If we get here, the user did not supply a connection string, build it
            connection_string = "jdbc:sqlserver://{host}:{port};databaseName={database}".format(
                host=self.config["database_host"],
                port=self.config["database_port"],
                database=self.config["database_name"]
            )
        self.logger.info(f"Using connection string {connection_string}")
        return connection_string

    def write_config(self, queries_to_run: List[QueryConfig]):
        """
        Writes the database configuration to a json file named after the endpoint ID
        """
        queries_json = [query.__dict__ for query in queries_to_run]
        config = {
            "connectionString": self.build_connection_string(),
            "username": self.config["database_user"],
            "queries": queries_json,
            "driverClassName": "com.microsoft.sqlserver.jdbc.SQLServerDriver",
        }

        path = Path(self.temp_folder, "config")
        if not path.exists():
            os.mkdir(path)
        with open(Path(path, f"{self.activation.entity_id}.json"), "w") as f:
            json.dump(config, f, indent=2)

    def read_results_from_file(self) -> Result:
        """
        Reads the <endpoint_id>.json file inside the results folder
        and parses the data into a models.Result object
        """
        results_file = Path(self.temp_folder, "results", f"{self.activation.entity_id}.json")
        if results_file.exists():
            with open(results_file, "r") as f:
                try:
                    return Result.from_json(json.load(f))
                except Exception as e:
                    self.logger.error(f"Could not parse json from {results_file}: {e}")
                    raise NothingToReportException(f"Error: {e}")
        return Result()

    def report_metric(self, key: str, value: Any, dimensions: Optional[Dict[str, str]] = None, metric_type: str = "absolute"):
        # Get the method based on the name of the metric type
        # Ie, 'absolute' gives us the custom_device.absolute method
        method: Callable = self.custom_device.__getattribute__(metric_type)
        self.logger.debug(f"Sending {metric_type} metric with key={key}, value={value}, dimensions={dimensions}")
        method(key, value, dimensions)

    def parse_logs(self):
        """
        Simply logs all lines from the java process logs to the extension logs
        """
        log = Path(self.temp_folder, "log", f"{self.activation.entity_id}.log")
        cache = Path(self.temp_folder, f"{self.activation.entity_id}.json")

        for line in LogProxy(log, cache).get_lines_from_file():
            self.logger.info(line)


def create_or_get_temp_folder() -> Path:
    temp_folder = Path(tempfile.gettempdir(), "dynatrace-mssql-extension")
    if not temp_folder.exists():
        os.mkdir(temp_folder)
    return temp_folder
